/* tcp port forward utility
Copyright (c) 2003 - 2005 Heikki Orsila <heikki.orsila@iki.fi>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <netinet/in.h>

#include "net.h"
#include "pipe.h"
#include "accounting.h"

static struct net_host_id *allowed_hosts = NULL;
static int n_allowed_hosts = 0;

static char *accounting_file = NULL;


static int get_host_port(char **address, char **port, char *in)
{
  char *sep;
  char *nstr;

  if (!(nstr = strdup(in)))
    return 0;

  if (!(sep = strrchr(nstr, (int) ':'))) {
    return 0;
  }

  *sep = 0;

  *address = strdup(nstr);
  *port = strdup(sep + 1);

  free (nstr);

  if (!(*address) || !(*port)) {
    free (*address);
    free (*port);
    return 0;
  }

  return 1;
}


static void pf_handle_client(int src_fd, char *dst_host, char *dst_port)
{
  int dst_fd;
  struct pipe p1;
  struct pipe p2;
  char buf[1024];

  if (accounting_file) {
    get_account(buf, sizeof(buf), accounting_file);
    dst_host = buf;
  }

  if ((dst_fd = net_open(dst_host, dst_port, "tcp", 0)) < 0) {
    fprintf(stderr, "tcp-pf: couldn't connect to %s:%s\n", dst_host, dst_port);
    return;
  }

  memset(&p1, 0, sizeof(struct pipe));
  memset(&p2, 0, sizeof(struct pipe));

  p1.src_fd = src_fd;
  p1.dst_fd = dst_fd;
  p2.src_fd = dst_fd;
  p2.dst_fd = src_fd;

  /* here is the main loop copying data from one socket to another */
  pmr_loop(&p1, &p2, dst_host, accounting_file);

  if (accounting_file != NULL) {
    account_increment(dst_host, accounting_file, p1.accbytes + p2.accbytes);
    p1.accbytes = p2.accbytes = 0;
  }

  close(p1.src_fd);
  close(p1.dst_fd);
  close(p2.src_fd);
  close(p2.dst_fd);
}


static int pf_accept_loop(char *src_port, char *dst_host, char *dst_port)
{
  int server_fd;
  int src_fd;
  pid_t child;

  char sockbuf[256];
  struct sockaddr_in *sockname = (struct sockaddr_in *) sockbuf;
  int socklen;
  int i;
  struct net_host_id id;
  char printed_id[256];
  int got_info;

  if ((server_fd = net_bind(0, 0, src_port, "tcp", 1 /* REUSE */)) < 0) {
    fprintf(stderr, "tcp-pf: can't bind server port\n");
    exit(-1);
  }

  if (listen(server_fd, 5)) {
    perror("can't listen on server port");
    exit(-1);
  }

  while (1) {
    socklen = sizeof(sockbuf);
    src_fd = accept(server_fd, (struct sockaddr *) sockname, &socklen);
    if (src_fd < 0) {
      if (errno == EAGAIN || errno == EINTR) {
	continue;
      }
      perror("client accept failed");
      goto err_out;
    }

    if (!(got_info = net_get_id(&id, (struct sockaddr *) sockname))) {
      fprintf(stderr, "client connected. can't get info on socket. rejecting.\n");
      if (n_allowed_hosts > 0) {
	goto reject;
      }
    }

    if (got_info) {
      printed_id[0] = 0;
      net_print_id(printed_id, &id, sizeof(printed_id));

      if (n_allowed_hosts > 0) {
	for (i = 0; i < n_allowed_hosts; i++) {
	  if (net_are_same_ips(&id, allowed_hosts + i))
	    goto accept;
	}
	fprintf(stderr, "client connected from %s => reject\n", printed_id);
	goto reject;
      }
    accept:
      fprintf(stderr, "client connected from %s => accept\n", printed_id);
    }

    child = fork();
    if (!child) {
      close(server_fd);
      child = fork();
      if (!child) {
	pf_handle_client(src_fd, dst_host, dst_port);
	printf("client handled\n");
      } else if (child == -1) {
	perror("children can not reproduce");
      }
      abort();

    } else if (child == -1) {
      perror("can not have children. client will die.");

    } else {
      wait(0);
    }

  reject:

    close(src_fd);
  }

  return 1;

 err_out:
  close(server_fd);
  return 0;
}


static void print_help(void)
{
  printf("\n");
  printf("tcp-pf %s by Heikki Orsila <heikki.orsila@iki.fi>\n", VERSION);
  printf("\n");
  printf("OPTIONS\n");
  printf("\n");
  printf(" -a/--allow host   allow incoming connections from host. this option can be\n");
  printf("                   used many times to select many hosts\n");
  printf(" -d host:port      forward incoming connections to host:port\n");
  printf(" -h/--help         this crap\n");
  printf(" --lazy            let tcp connection hang until both ends have shutdown\n");
  printf(" -l/--list file    multiplex target connections to hosts listed in file\n");
  printf("                   according to transfer history. the host that has least\n");
  printf("                   transfered bytes is selected. (network load balancing)\n");
  printf(" -p/--dport port   set destination port (use with -l)\n");
  printf(" -r/--rate rate    maximum number of bytes per second going through one\n");
  printf("                   connection.\n");
  printf(" -s port           set source port to accept connections\n");
}


int main(int argc, char **argv)
{
  int i;
  char *src_port = NULL;
  char *dst_host = NULL;
  char *dst_port = NULL;
  struct net_host_id id;
  struct net_host_id *new_allowed;

  for (i = 1; i < argc;) {

    if (!strcmp("-s", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give source port.\n");
	print_help();
	return -1;
      }
      if (!(src_port = strdup(argv[i + 1]))) {
	fprintf(stderr, "not enough memory for src_port\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("-d", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give destination host:port\n");
	print_help();
	return -1;
      }
      if (!get_host_port(&dst_host, &dst_port, argv[i + 1])) {
	return -1;
      }
      i += 2;

    } else if (!strcmp("--list", argv[i]) || !strcmp("-l", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give accounting file\n");
	print_help();
	return -1;
      }
      accounting_file = strdup(argv[i + 1]);
      if (accounting_file == NULL) {
	fprintf(stderr, "not enough memory for accounting filename\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("--dport", argv[i]) || !strcmp("-p", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give dst port.\n");
	print_help();
	return -1;
      }
      if ((dst_port = strdup(argv[i + 1])) == NULL) {
	fprintf(stderr, "not enough memory for dst port\n");
	return -1;
      }
      i += 2;

    } else if (!strcmp("--lazy", argv[i])) {
      kill_on_eof = 0;
      i++;

    } else if (!strcmp("--allow", argv[i]) || !strcmp("-a", argv[i])) {
      if ((i + 1) >= argc) {
	fprintf(stderr, "not enough args. give allowed hostname.\n");
	print_help();
	return -1;
      }
      if (!net_info(&id, argv[i + 1])) {
	fprintf(stderr, "nothing known about %s\n", argv[i + 1]);
	return -1;
      }
      new_allowed = realloc(allowed_hosts, (n_allowed_hosts + 1) * sizeof(struct net_host_id));
      if (!new_allowed) {
	fprintf(stderr, "no memory for allowed hosts\n");
	return -1;
      }
      allowed_hosts = new_allowed;
      allowed_hosts[n_allowed_hosts] = id;
      n_allowed_hosts++;
      i += 2;

    } else if (!strcmp("-r", argv[i]) || !strcmp("--rate", argv[i])) {
      if ((i + 1) < argc) {
        pmr_max_rate = atoi(argv[i + 1]);
	if (pmr_max_rate <= 0) {
	  fprintf (stderr, "illegal bytes per second value (%d)\n", pmr_max_rate);
	  return -1;
	}
      } else {
        fprintf (stderr, "expecting a rate limit value (bytes per second) with -r\n");
	print_help();
        return -1;
      }
      i += 2;

    } else if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      print_help();
      return 0;

    } else {
      fprintf(stderr, "illegal args (%s)\n", argv[i]);
      print_help();
      return -1;
    }
  }

  if (src_port == NULL) {
    fprintf(stderr, "you must provide source port with -s port!\n");
    print_help();
    return -1;
  }

  if (accounting_file == NULL) {
    if (dst_host == NULL || dst_port == NULL) {
      fprintf(stderr, "you must provide destination host:port with -d host:port, or alternatively\n");
      fprintf(stderr, "you can provide destination list and destination port with -l listfile -p port\n");
      print_help();
      return -1;
    }
  } else if (dst_port == NULL) {
    fprintf(stderr, "you must provide destinatio port with -p port\n");
    print_help();
    return -1;
  }

  return pf_accept_loop(src_port, dst_host, dst_port) ? -1 : 0;
}
